/* Gnome-Look theme engine drawing primitives
 *
 * Copyright Duncan Pierce 1999, based upon an early version of
 * ThinIce theme by Tim Gerla <timg@means.net> and Tomas gren
 * <stric@ing.umu.se> (in turn based upon original code by Rasterman)
 * Tim and Tomas and me thank Evan Lawrence for the option code.
 * -- not sure how much ThinIce is left in here now.
 *
 * GPL license
 *
 * This has to be the most arbitrary interface I've ever seen.  If
 * you're listening out there - don't write a Gtk theme until someone
 * has cleaned this interface up a *lot*.  Most of it doesn't get
 * called, except for `draw_box' which basically does practically
 * everything.  Ouch.  Also, it doesn't know whether it wants to be
 * drawing "physical" things like boxes and lines, or "logical" things
 * like sliders and handles.
 *
 *
 * I wish this worked (in draw_box)...
 *   if (DETAIL("trough"))
 *   {
 *     if (width > height)
 *       draw_hline (style, window, state_type, area, widget, "trough",
 *                   x, x + width - 1, y + (height / 2) - 1);
 *     else
 *       draw_vline (style, window, state_type, area, widget, "trough",
 *                   y, y + height - 1, x + (width / 2) - 1);
 *   }
 *
 * News flash!  I now know why this doesn't work - it is a deliberate
 * "bug" in GtkRange, where the author optimised the redrawing (by not
 * doing any).  gtkrange.c:gtk_range_expose is the place.  Must fix.
 *
 * There is a lot of off-by-one in Gtk-1.2's calls into this
 * interface.  For example, draw_focus interprets width and height of
 * the focused area differently to, say, draw_box.  Even better,
 * widgets disagree about their meaning too, so you cannot easily
 * compensate for this.  Check out range widgets and compare to
 * entries.  I'd like to vote for the range widget way of doing it,
 * since that is consistent with other interpretations of width and
 * height.  But this means loads of things are off, so... just bodge
 * it.
 *
 * I should look at `gtkstyle.c':2229 for flat_box stuff.
 * */

#include "gnomelook_theme.h"
#include <math.h>
#include <stdarg.h>

#define DETAIL(xx) ((detail) && (!strcmp(xx, detail)))


static void draw_hline (GtkStyle * style,
                        GdkWindow * window,
                        GtkStateType state_type,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        gint x1,
                        gint x2,
                        gint y);

static void draw_vline (GtkStyle * style,
                        GdkWindow * window,
                        GtkStateType state_type,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        gint y1,
                        gint y2,
                        gint x);

static void draw_shadow (GtkStyle * style,
                         GdkWindow * window,
                         GtkStateType state_type,
                         GtkShadowType shadow_type,
                         GdkRectangle * area,
                         GtkWidget * widget,
                         gchar * detail,
                         gint x,
                         gint y,
                         gint width,
                         gint height);

static void draw_polygon (GtkStyle * style,
                          GdkWindow * window,
                          GtkStateType state_type,
                          GtkShadowType shadow_type,
                          GdkRectangle * area,
                          GtkWidget * widget,
                          gchar * detail,
                          GdkPoint * point,
                          gint npoints,
                          gint fill);

static void draw_arrow (GtkStyle * style,
                        GdkWindow * window,
                        GtkStateType state_type,
                        GtkShadowType shadow_type,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        GtkArrowType arrow_type,
                        gint fill,
                        gint x,
                        gint y,
                        gint width,
                        gint height);

static void draw_diamond (GtkStyle * style,
                          GdkWindow * window,
                          GtkStateType state_type,
                          GtkShadowType shadow_type,
                          GdkRectangle * area,
                          GtkWidget * widget,
                          gchar * detail,
                          gint x,
                          gint y,
                          gint width,
                          gint height);

static void draw_oval (GtkStyle * style,
                       GdkWindow * window,
                       GtkStateType state_type,
                       GtkShadowType shadow_type,
                       GdkRectangle * area,
                       GtkWidget * widget,
                       gchar * detail,
                       gint x,
                       gint y,
                       gint width,
                       gint height);

static void draw_string (GtkStyle * style,
                         GdkWindow * window,
                         GtkStateType state_type,
                         GdkRectangle * area,
                         GtkWidget * widget,
                         gchar * detail,
                         gint x,
                         gint y,
                         const gchar * string);

static void draw_box (GtkStyle * style,
                      GdkWindow * window,
                      GtkStateType state_type,
                      GtkShadowType shadow_type,
                      GdkRectangle * area,
                      GtkWidget * widget,
                      gchar * detail,
                      gint x,
                      gint y,
                      gint width,
                      gint height);

static void draw_flat_box (GtkStyle * style,
                           GdkWindow * window,
                           GtkStateType state_type,
                           GtkShadowType shadow_type,
                           GdkRectangle * area,
                           GtkWidget * widget,
                           gchar * detail,
                           gint x,
                           gint y,
                           gint width,
                           gint height);

static void draw_check (GtkStyle * style,
                        GdkWindow * window,
                        GtkStateType state_type,
                        GtkShadowType shadow_type,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        gint x,
                        gint y,
                        gint width,
                        gint height);

static void draw_option (GtkStyle * style,
                         GdkWindow * window,
                         GtkStateType state_type,
                         GtkShadowType shadow_type,
                         GdkRectangle * area,
                         GtkWidget * widget,
                         gchar * detail,
                         gint x,
                         gint y,
                         gint width,
                         gint height);

static void draw_cross (GtkStyle * style,
                        GdkWindow * window,
                        GtkStateType state_type,
                        GtkShadowType shadow_type,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        gint x,
                        gint y,
                        gint width,
                        gint height);

static void draw_ramp (GtkStyle * style,
                       GdkWindow * window,
                       GtkStateType state_type,
                       GtkShadowType shadow_type,
                       GdkRectangle * area,
                       GtkWidget * widget,
                       gchar * detail,
                       GtkArrowType arrow_type,
                       gint x,
                       gint y,
                       gint width,
                       gint height);

static void draw_tab (GtkStyle * style,
                      GdkWindow * window,
                      GtkStateType state_type,
                      GtkShadowType shadow_type,
                      GdkRectangle * area,
                      GtkWidget * widget,
                      gchar * detail,
                      gint x,
                      gint y,
                      gint width,
                      gint height);

static void draw_shadow_gap (GtkStyle * style,
                             GdkWindow * window,
                             GtkStateType state_type,
                             GtkShadowType shadow_type,
                             GdkRectangle * area,
                             GtkWidget * widget,
                             gchar * detail,
                             gint x,
                             gint y,
                             gint width,
                             gint height,
                             GtkPositionType gap_side,
                             gint gap_x,
                             gint gap_width);

static void draw_box_gap (GtkStyle * style,
                          GdkWindow * window,
                          GtkStateType state_type,
                          GtkShadowType shadow_type,
                          GdkRectangle * area,
                          GtkWidget * widget,
                          gchar * detail,
                          gint x,
                          gint y,
                          gint width,
                          gint height,
                          GtkPositionType gap_side,
                          gint gap_x,
                          gint gap_width);

static void draw_extension (GtkStyle * style,
                            GdkWindow * window,
                            GtkStateType state_type,
                            GtkShadowType shadow_type,
                            GdkRectangle * area,
                            GtkWidget * widget,
                            gchar * detail,
                            gint x,
                            gint y,
                            gint width,
                            gint height,
                            GtkPositionType gap_side);

static void draw_focus (GtkStyle * style,
                        GdkWindow * window,
                        GdkRectangle * area,
                        GtkWidget * widget,
                        gchar * detail,
                        gint x,
                        gint y,
                        gint width,
                        gint height);

static void draw_slider (GtkStyle * style,
                         GdkWindow * window,
                         GtkStateType state_type,
                         GtkShadowType shadow_type,
                         GdkRectangle * area,
                         GtkWidget * widget,
                         gchar * detail,
                         gint x,
                         gint y,
                         gint width,
                         gint height,
                         GtkOrientation orientation);

static void draw_handle (GtkStyle * style,
                         GdkWindow * window,
                         GtkStateType state_type,
                         GtkShadowType shadow_type,
                         GdkRectangle * area,
                         GtkWidget * widget,
                         gchar * detail,
                         gint x,
                         gint y,
                         gint width,
                         gint height,
                         GtkOrientation orientation);


GtkStyleClass gnomelook_default_class =
{
  2,
  2,
  draw_hline,
  draw_vline,
  draw_shadow,
  draw_polygon,
  draw_arrow,
  draw_diamond,
  draw_oval,
  draw_string,
  draw_box,
  draw_flat_box,
  draw_check,
  draw_option,
  draw_cross,
  draw_ramp,
  draw_tab,
  draw_shadow_gap,
  draw_box_gap,
  draw_extension,
  draw_focus,
  draw_slider,
  draw_handle
};




/*********************************************************************/

/* Some stuff I have yet to make use of... */
typedef struct {
  int top_left : 1;
  int top : 1; 
  int top_right : 1;
  int right : 1;
  int bottom_right : 1;
  int bottom : 1;
  int bottom_left : 1;
  int left : 1;
} FrameDetails;

static void
draw_generic_frame (GtkStyle *style,
                    GdkWindow *window,
                    GtkStateType state_type,
                    GdkRectangle *area,
                    GtkWidget *widget,
                    gchar *detail,
                    gint x,
                    gint y,
                    gint width,
                    gint height,
                    gint rounded,
                    FrameDetails frame_details)
{
}

static void
style_set_clip (GtkStyle *style,
                GtkStateType state_type,
                GdkRectangle *area)
{
  gdk_gc_set_clip_rectangle (style->white_gc, area);
  gdk_gc_set_clip_rectangle (style->black_gc, area);
  gdk_gc_set_clip_rectangle (style->light_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->mid_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->dark_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->base_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->fg_gc[state_type], area);
  gdk_gc_set_clip_rectangle (style->text_gc[state_type], area);
}

static void
style_set_shadow_colours (GtkStyle *style,
                          GtkStateType state_type,
                          GtkShadowType shadow_type,
                          GdkGC **gc1,
                          GdkGC **gc2)
{
  if (shadow_type == GTK_SHADOW_OUT || shadow_type == GTK_SHADOW_ETCHED_OUT)
    {
      *gc1 = style->light_gc[state_type];
      *gc2 = style->dark_gc[state_type];
    }
  else
    {
      *gc1 = style->dark_gc[state_type];
      *gc2 = style->light_gc[state_type];
    }
}

                   

/*********************************************************************/


static void
draw_hline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x1,
	   gint x2,
	   gint y)
{
  gint                thickness_light;
  gint                thickness_dark;
  gint                i;
  GdkGC *light_gc;
  GdkGC *dark_gc;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  /* Suppress drawing those irritating lines on panes. */
  if (DETAIL ("vpaned"))
    return;

  thickness_light = style->klass->ythickness / 2;
  thickness_dark = style->klass->ythickness - thickness_light;

  /* Handles and sliders go out instead of in. */
  if (DETAIL ("handle") || DETAIL ("sliderknurl"))
    {
      light_gc = style->dark_gc[state_type];
      dark_gc = style->light_gc[state_type];
    }
  else
    {
      light_gc = style->light_gc[state_type];
      dark_gc = style->dark_gc[state_type];
    }

  if (area)
  {
    gdk_gc_set_clip_rectangle(light_gc, area);
    gdk_gc_set_clip_rectangle(dark_gc, area);
  }

  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line(window, light_gc, x2 - i - 1, y + i, x2, y + i);
    gdk_draw_line(window, dark_gc, x1, y + i, x2 - i - 1, y + i);
  }

  y += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line(window, dark_gc, x1, y + i, x1 + thickness_light - i - 1, y + i);
    gdk_draw_line(window, light_gc, x1 + thickness_light - i, y + i, x2, y + i);
  }

  if (area)
  {
    gdk_gc_set_clip_rectangle(light_gc, NULL);
    gdk_gc_set_clip_rectangle(dark_gc, NULL);
  }
}

static void
draw_vline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint y1,
	   gint y2,
	   gint x)
{
  gint thickness_light;
  gint thickness_dark;
  gint i;
  GdkGC *light_gc;
  GdkGC *dark_gc;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  /* Suppress drawing those irritating lines on panes. */
  if (DETAIL ("hpaned"))
    return;

  thickness_light = style->klass->xthickness / 2;
  thickness_dark = style->klass->xthickness - thickness_light;

  /* Handles and sliders go out instead of in. */
  if (DETAIL ("handle") || DETAIL ("sliderknurl"))
    {
      light_gc = style->dark_gc[state_type];
      dark_gc = style->light_gc[state_type];
    }
  else
    {
      light_gc = style->light_gc[state_type];
      dark_gc = style->dark_gc[state_type];
    }

  if (area)
  {
    gdk_gc_set_clip_rectangle(light_gc, area);
    gdk_gc_set_clip_rectangle(dark_gc, area);
  }

  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line(window, light_gc, x + i, y2 - i - 1, x + i, y2);
    gdk_draw_line(window, dark_gc, x + i, y1, x + i, y2 - i - 1);
  }

  x += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line(window, dark_gc, x + i, y1, x + i, y1 + thickness_light - i);
    gdk_draw_line(window, light_gc, x + i, y1 + thickness_light - i, x + i, y2);
  }

  if (area)
  {
    gdk_gc_set_clip_rectangle(light_gc, NULL);
    gdk_gc_set_clip_rectangle(dark_gc, NULL);
  }
}

static void
draw_shadow(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
  GdkGC *gc1;
  GdkGC *gc2;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  if (shadow_type == GTK_SHADOW_IN || shadow_type == GTK_SHADOW_ETCHED_IN)
    {
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
    }
  else
    {
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);
    }

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      gdk_draw_rectangle (window, style->bg_gc[state_type], FALSE,
                          x, y, width - 1, height - 1);
      break;

    /* draw inner box... */
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
	gdk_draw_line(window, gc1,
		      x + 1, y + 1, x + width-3, y + 1);
	gdk_draw_line(window, gc1,
		      x + 1, y + 1, x + 1, y + height-3);
	
	gdk_draw_line(window, gc2,
		      x + 1, y + height-2, x + width-2, y + height-2);
	gdk_draw_line(window, gc2,
		      x + width-2, y + 1, x + width-2, y + height-3);

    /* continue into GTK_SHADOW_IN to draw outer box... */
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
	gdk_draw_line(window, gc2,
		      x, y, x + width-1, y);
	gdk_draw_line(window, gc2,
		      x, y, x, y + height-1);
	
	gdk_draw_line(window, gc1,
		      x, y + height-1, x + width-1, y + height-1);
	gdk_draw_line(window, gc1,
		      x + width-1, y, x + width-1, y + height-1);
    }

  /* if it wasn't an ETCHED_* shadow we still have the inner box to draw */
  if (shadow_type != GTK_SHADOW_ETCHED_IN
      && shadow_type != GTK_SHADOW_ETCHED_OUT)
    gdk_draw_rectangle (window, style->bg_gc[state_type], FALSE,
                        x + 1, y + 1, width - 3, height - 3);

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], NULL);
    }
}

static void
draw_polygon(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     GdkPoint * points,
	     gint npoints,
	     gint fill)
{
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif /* M_PI */
#ifndef M_PI_4
#define M_PI_4  0.78539816339744830962
#endif /* M_PI_4 */

  static const gdouble pi_over_4 = M_PI_4;
  static const gdouble pi_3_over_4 = M_PI_4 * 3;

  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;
  GdkGC              *gc4;
  gdouble             angle;
  gint                xadjust;
  gint                yadjust;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  g_return_if_fail(points != NULL);

  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
      gc3 = style->light_gc[state_type];
      gc4 = style->dark_gc[state_type];
      break;
    case GTK_SHADOW_ETCHED_IN:
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
      gc3 = style->dark_gc[state_type];
      gc4 = style->light_gc[state_type];
      break;
    case GTK_SHADOW_OUT:
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
      gc3 = style->dark_gc[state_type];
      gc4 = style->light_gc[state_type];
      break;
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
      gc3 = style->light_gc[state_type];
      gc4 = style->dark_gc[state_type];
      break;
    default:
      return;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      gdk_gc_set_clip_rectangle(gc3, area);
      gdk_gc_set_clip_rectangle(gc4, area);
    }

  if (fill)
    gdk_draw_polygon(window, style->bg_gc[state_type], TRUE, points, npoints);

  npoints--;

  for (i = 0; i < npoints; i++)
    {
      if ((points[i].x == points[i + 1].x) &&
	  (points[i].y == points[i + 1].y))
	{
	  angle = 0;
	}
      else
	{
	  angle = atan2(points[i + 1].y - points[i].y,
			points[i + 1].x - points[i].x);
	}

      if ((angle > -pi_3_over_4) && (angle < pi_over_4))
	{
	  if (angle > -pi_over_4)
	    {
	      xadjust = 0;
	      yadjust = 1;
	    }
	  else
	    {
	      xadjust = 1;
	      yadjust = 0;
	    }

	  gdk_draw_line(window, gc1,
			points[i].x - xadjust, points[i].y - yadjust,
			points[i + 1].x - xadjust, points[i + 1].y - yadjust);
	  gdk_draw_line(window, gc3,
			points[i].x, points[i].y,
			points[i + 1].x, points[i + 1].y);
	}
      else
	{
	  if ((angle < -pi_3_over_4) || (angle > pi_3_over_4))
	    {
	      xadjust = 0;
	      yadjust = 1;
	    }
	  else
	    {
	      xadjust = 1;
	      yadjust = 0;
	    }

	  gdk_draw_line(window, gc4,
			points[i].x + xadjust, points[i].y + yadjust,
			points[i + 1].x + xadjust, points[i + 1].y + yadjust);
	  gdk_draw_line(window, gc2,
			points[i].x, points[i].y,
			points[i + 1].x, points[i + 1].y);
	}
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      gdk_gc_set_clip_rectangle(gc3, NULL);
      gdk_gc_set_clip_rectangle(gc4, NULL);
    }
}


static void
draw_arrow (GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            gchar * detail,
            GtkArrowType arrow_type,
            gint fill,
            gint x,
            gint y,
            gint width,
            gint height)
{
  GdkGC *bg_gc;
  GdkGC *outline_gc;
  gint half_width, half_height;
  gint size, offset;
  gint start, end, increment;
  GdkPoint points[3];

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  /* Scrollbar arrows have a button-style box drawn around them, and
   * arrow is made (even) smaller. */
  if (DETAIL ("hscrollbar") || DETAIL ("vscrollbar"))
    {
      draw_box (style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
      width -= 4;
      height -= 4;
      x += 2;
      y += 2;
    }
  else if (DETAIL ("menuitem"))
    {
      width -= 2;
      height -= 2;
      x += 1;
      y += 1;
    }

  /* Give up if too small. */
  if (width < 4 || height < 4)
    return;

  /* Make everything a nice size and adjust position of arrow, if
   * necessary. */
  if (height > width)
    {
      y += (height - width) / 2;
      height = width = (width - 1) | 1;
    }
  else
    {
      x += (width - height) / 2;
      height = width = (height - 1) | 1;
    }

  /* The usual GC stuff. */
  bg_gc = style->base_gc[state_type];
  outline_gc = style->black_gc;
  if (area)
    {
      gdk_gc_set_clip_rectangle (outline_gc, area);
      gdk_gc_set_clip_rectangle (bg_gc, area);
    }

  /* Set up for draw loop. */
  switch (arrow_type)
    {
    case GTK_ARROW_RIGHT:
      increment = +1;
      start = x;
      end = x + width;
      break;
    case GTK_ARROW_LEFT:
      increment = -1;
      start = x + width - 1;
      end = x - 1;
      break;
    case GTK_ARROW_DOWN:
      increment = +1;
      start = y;
      end = y + height;
      break;;
    case GTK_ARROW_UP:
      increment = -1;
      start = y + height - 1;
      end = y - 1;
      break;
    }

  /* Draw loops. */
  if (arrow_type == GTK_ARROW_LEFT || arrow_type == GTK_ARROW_RIGHT)
    {
      gdk_draw_line (window, outline_gc, start, y,
                     start, y + height - 1);
      for (start += increment; start != end;
           start += (increment << 1), y++, height -= 2)
        {
          if (fill)
            {
              gdk_draw_line (window, bg_gc,
                             start,              y + 1,
                             start,              y + height - 2);
              gdk_draw_line (window, bg_gc,
                             start + increment,  y + 1,
                             start + increment,  y + height - 2);
            }
          gdk_draw_line (window, outline_gc,
                         start,              y,
                         start + increment,  y);
          gdk_draw_line (window, outline_gc,
                         start,              y + height - 1,
                         start + increment,  y + height - 1);
        }
      gdk_draw_point (window, outline_gc, start, y);
    }
  else
    {
      gdk_draw_line (window, outline_gc, x, start,
                     x + width - 1, start);
      for (start += increment; start != end;
           start += (increment << 1), x++, width -= 2)
        {
          if (fill)
            {
              gdk_draw_line (window, bg_gc,
                             x + 1,          start,
                             x + width - 2,  start);
              gdk_draw_line (window, bg_gc,
                             x + 1,          start + increment,
                             x + width - 2,  start + increment);
            }
          gdk_draw_line (window, outline_gc,
                         x,              start,
                         x,              start + increment);
          gdk_draw_line (window, outline_gc,
                         x + width - 1,  start,
                         x + width - 1,  start + increment);
        }
      gdk_draw_point (window, outline_gc, x, start);
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (outline_gc, NULL);
      gdk_gc_set_clip_rectangle (bg_gc, NULL);
    }
}


static void
draw_diamond(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height)
{
  gint                half_width;
  gint                half_height;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  /* What does all this stuff do?  Anything? */
#if 0
  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  half_width = width / 2;
  half_height = height / 2;

  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->dark_gc[state_type], area);
      gdk_gc_set_clip_rectangle(style->black_gc, area);
    }
  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    case GTK_SHADOW_OUT:
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->dark_gc[state_type],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->light_gc[state_type],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[state_type],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->light_gc[state_type],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    default:
      break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->dark_gc[state_type], NULL);
      gdk_gc_set_clip_rectangle(style->black_gc, NULL);
    }
#endif
}

static void
draw_oval (GtkStyle * style,
           GdkWindow * window,
           GtkStateType state_type,
           GtkShadowType shadow_type,
           GdkRectangle * area,
           GtkWidget * widget,
           gchar * detail,
           gint x,
           gint y,
           gint width,
           gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
}

static void
draw_string (GtkStyle * style,
             GdkWindow * window,
             GtkStateType state_type,
             GdkRectangle * area,
             GtkWidget * widget,
             gchar * detail,
             gint x,
             gint y,
             const gchar * string)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if (area)
    gdk_gc_set_clip_rectangle (style->text_gc[state_type], area);
  gdk_draw_string (window, style->font,
                   style->text_gc[state_type], 
                   x, y, string);
  if (area)
    gdk_gc_set_clip_rectangle (style->text_gc[state_type], NULL);
}

static void
draw_box(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
  GdkGC *bg_gc;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  if (DETAIL ("buttondefault"))
    state_type = GTK_STATE_INSENSITIVE;

  bg_gc = style->bg_gc[state_type];

  if (area)
    gdk_gc_set_clip_rectangle (bg_gc, area);
  gdk_draw_rectangle (window, bg_gc, TRUE, x, y, width, height);
  if (area)
    gdk_gc_set_clip_rectangle (bg_gc, NULL);

  /* Draw menus with a black outline (similar to tooltips) - this
   * helps them stand out when popped up.  It doesn't look quite so
   * good on tear-off menus, but I don't even want to think about how
   * much effort would be involved in fixing this. */
  if (DETAIL ("menu"))
    {
      if (area)
        gdk_gc_set_clip_rectangle (style->black_gc, area);

      gdk_draw_rectangle (window, style->black_gc, FALSE, x, y,
                          width - 1, height - 1);
      x++;
      y++;
      width -= 2;
      height -= 2;

      if (area)
        gdk_gc_set_clip_rectangle (style->black_gc, NULL);
      /* FIXME: another draw-too-much problem. */
    }

  /* Nasty kludge: we don't know the orientation of a scrollbar, so we
   * have to resort to arbitrary means to determine it.  This code is
   * a compromise that usually works well.  Scrollbars sliders ought
   * to be drawn with `draw_slider', but aren't. */
  else if (DETAIL ("slider") && (width > 16 || height > 16))
    {
      if (height >= width)
        draw_slider (style, window, state_type, shadow_type, area, widget,
                     "scrollbar", x, y, width, height,
                     GTK_ORIENTATION_VERTICAL);
      else
        draw_slider (style, window, state_type, shadow_type, area, widget,
                     "scrollbar", x, y, width, height,
                     GTK_ORIENTATION_HORIZONTAL);
      return;
    }

  
  if (! DETAIL ("menuitem"))
    gtk_paint_shadow (style, window, state_type, shadow_type, area, widget,
                      detail, x, y, width, height);
}



static void
draw_flat_box (GtkStyle * style,
               GdkWindow * window,
               GtkStateType state_type,
               GtkShadowType shadow_type,
               GdkRectangle * area,
               GtkWidget * widget,
               gchar * detail,
               gint x,
               gint y,
               gint width,
               gint height)
{
  GdkGC *gc;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  gc = style->bg_gc[state_type];
  
  if (area)
    gdk_gc_set_clip_rectangle (gc, area);
  
  gdk_draw_rectangle (window, gc, TRUE, x, y, width, height);

  if (DETAIL("tooltip"))
    {
      /* FIXME: this is a slight kludge - it draws the border too
         thick (not visible). */
      draw_shadow (style, window, state_type, GTK_SHADOW_OUT, area,
                   widget, detail, x + 1, y + 1, width - 2, height - 2);
      gdk_draw_rectangle (window, style->black_gc, FALSE,
                          x, y, width - 1, height - 1);
    }

  if (area)
    gdk_gc_set_clip_rectangle (gc, NULL);
}

static void
draw_check(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
    gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		  x, y, width, height);
}

static void
draw_option(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            gchar * detail,
            gint x,
            gint y,
            gint width,
            gint height)
{
  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;

  if (shadow_type == GTK_SHADOW_IN)
    {
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
    }
  else
    {
      gc1 = style->light_gc[state_type]; 
      gc2 = style->dark_gc[state_type];
    }
  
  gc3 = style->bg_gc[state_type];
 
  if (area)                           
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      gdk_gc_set_clip_rectangle(gc3, area);
    }
  
  gdk_draw_arc(window, gc3, TRUE, x + 1, y + 1, width - 2, height - 2, 0, 360 *
	       64);
  gdk_draw_arc(window, gc1, FALSE, x, y, width, height, 45 * 64, 225 * 64);
  gdk_draw_arc(window, gc2, FALSE, x, y, width, height, 225 * 64, 180 * 64);
  
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      gdk_gc_set_clip_rectangle(gc3, NULL);
    }
}


static void
draw_cross(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

static void
draw_ramp (GtkStyle * style,
           GdkWindow * window,
           GtkStateType state_type,
           GtkShadowType shadow_type,
           GdkRectangle * area,
           GtkWidget * widget,
           gchar * detail,
           GtkArrowType arrow_type,
           gint x,
           gint y,
           gint width,
           gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

static void
draw_tab (GtkStyle * style,
          GdkWindow * window,
          GtkStateType state_type,
          GtkShadowType shadow_type,
          GdkRectangle * area,
          GtkWidget * widget,
          gchar * detail,
          gint x,
          gint y,
          gint width,
          gint height)
{
  GdkGC *bg_gc;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if (DETAIL ("optionmenutab"))
    {
      bg_gc = style->base_gc[state_type];

      if (area)
        {
          gdk_gc_set_clip_rectangle (style->black_gc, area);
          gdk_gc_set_clip_rectangle (bg_gc, area);
        }

      gdk_draw_rectangle (window, bg_gc, TRUE, x, y, width, height);
      gdk_draw_rectangle (window, style->black_gc, FALSE, x, y,
                          width - 1, height - 1);

      if (area)
        {
          gdk_gc_set_clip_rectangle (style->black_gc, NULL);
          gdk_gc_set_clip_rectangle (bg_gc, NULL);
        }
    }
  else
    gtk_paint_box (style, window, state_type, shadow_type, area, widget,
                   detail, x, y, width, height);
}


static void
draw_shadow_gap(GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		gchar * detail,
		gint x,
		gint y,
		gint width,
		gint height,
		GtkPositionType gap_side,
		gint gap_x,
		gint gap_width)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);


  gtk_paint_shadow(style, window, state_type, shadow_type, area,
                   widget, detail,  x,y,  width,height);


  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    }

  gtk_style_apply_default_pixmap(style, window, state_type, area,
				 rect.x, rect.y, rect.width, rect.height);
}

static void
draw_box_gap(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height,
	     GtkPositionType gap_side,
	     gint gap_x,
	     gint gap_width)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    }

  gtk_style_apply_default_pixmap(style, window, state_type, area,
				 rect.x, rect.y, rect.width, rect.height);
}

static void
draw_extension (GtkStyle * style,
                GdkWindow * window,
                GtkStateType state_type,
                GtkShadowType shadow_type,
                GdkRectangle * area,
                GtkWidget * widget,
                gchar * detail,
                gint x,
                gint y,
                gint width,
                gint height,
                GtkPositionType gap_side)
{
  GdkGC *gc1;
  GdkGC *gc2;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  /* NB notebook tabs are detail "tab" */

  if (shadow_type == GTK_SHADOW_IN || shadow_type == GTK_SHADOW_ETCHED_IN)
    {
      gc1 = style->light_gc[state_type];
      gc2 = style->dark_gc[state_type];
    }
  else
    {
      gc1 = style->dark_gc[state_type];
      gc2 = style->light_gc[state_type];
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, area);
      gdk_gc_set_clip_rectangle (gc2, area);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);
    }

  /* Draw background. */
  gdk_draw_rectangle (window, style->bg_gc[state_type], TRUE,
                      x, y, width, height);

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      gdk_draw_rectangle (window, style->bg_gc[state_type], FALSE,
                          x, y, width - 1, height - 1);
      break;

    /* draw inner box... */
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
        if (gap_side != GTK_POS_TOP)
          gdk_draw_line(window, gc1,
                        x + 1, y + 1, x + width-3, y + 1);
        if (gap_side != GTK_POS_LEFT)
          gdk_draw_line(window, gc1,
                        x + 1, y + 1, x + 1, y + height-3);
        if (gap_side != GTK_POS_BOTTOM)
          gdk_draw_line(window, gc2,
                        x + 1, y + height-2, x + width-2, y + height-2);
        if (gap_side != GTK_POS_RIGHT)
          gdk_draw_line(window, gc2,
                        x + width-2, y + 1, x + width-2, y + height-3);

    /* continue into GTK_SHADOW_IN to draw outer box... */
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
        if (gap_side != GTK_POS_TOP)
          gdk_draw_line(window, gc2,
                        x, y, x + width-1, y);
        if (gap_side != GTK_POS_LEFT)
          gdk_draw_line(window, gc2,
                        x, y, x, y + height-1);
        if (gap_side != GTK_POS_BOTTOM)
          gdk_draw_line(window, gc1,
                        x, y + height-1, x + width-1, y + height-1);
        if (gap_side != GTK_POS_RIGHT)
          gdk_draw_line(window, gc1,
                        x + width-1, y, x + width-1, y + height-1);
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle (gc1, NULL);
      gdk_gc_set_clip_rectangle (gc2, NULL);
      gdk_gc_set_clip_rectangle (style->bg_gc[state_type], NULL);
    }
}

static void
draw_focus(GtkStyle * style,
	   GdkWindow * window,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  GdkGC *gc;
  gchar dashes[2] = {4, 4};
  GdkPoint points[4];

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  if (DETAIL ("trough"))
    {
      width--;
      height--;
    }

  points[0].x = x;
  points[0].y = y;
  points[1].x = x + width;
  points[1].y = y;
  points[2].x = x + width;
  points[2].y = y + height;
  points[3].x = x;
  points[3].y = y + height;

  gc = gdk_gc_new (window);

  /* Draw black dashes. */
  gdk_gc_copy (gc, style->black_gc);
  gdk_gc_set_dashes (gc, 0, dashes, 2);
  gdk_gc_set_line_attributes (gc, 1, GDK_LINE_ON_OFF_DASH,
                              GDK_CAP_BUTT, GDK_JOIN_MITER);
  if (area)
    gdk_gc_set_clip_rectangle (gc, area);
  gdk_draw_polygon (window, gc, FALSE, points, 4);

#if 0
  /* Draw white dashes. */
  gdk_gc_copy (gc, style->white_gc);
  gdk_gc_set_dashes (gc, 4, dashes, 2);
  gdk_gc_set_line_attributes (gc, 1, GDK_LINE_ON_OFF_DASH,
                              GDK_CAP_BUTT, GDK_JOIN_MITER);
  if (area)
    gdk_gc_set_clip_rectangle (gc, area);
  gdk_draw_polygon (window, gc, FALSE, points, 4);
#endif

  gdk_gc_unref (gc);
}

static void
draw_slider_knurl (GtkStyle *style,
                   GdkWindow *window,
                   GtkStateType state_type,
                   GdkRectangle *area,
                   GtkWidget *widget,
                   gint x,
                   gint y,
                   gint length,
                   gint num_lines,
                   GtkOrientation orientation)
{
  gint i;

  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    {
      /* Apply an offset so that we draw lines around a centre point -
       * don't ask how it works, just be thankful that it does. */
      x -= ((num_lines & (~1)) << 1) - (((~num_lines) & 1) << 1);

      for (i = 0; i < num_lines; i++, x += 4)
        draw_vline (style, window, state_type, area, widget, "sliderknurl",
                    y, y + length - 1, x);
    }
  else
    {
      y -= ((num_lines & (~1)) << 1) - (((~num_lines) & 1) << 1);

      for (i = 0; i < num_lines; i++, y += 4)
        draw_hline (style, window, state_type, area, widget, "sliderknurl",
                    x, x + length - 1, y);
    }
}

static void
draw_slider (GtkStyle * style,
             GdkWindow * window,
             GtkStateType state_type,
             GtkShadowType shadow_type,
             GdkRectangle * area,
             GtkWidget * widget,
             gchar * detail,
             gint x,
             gint y,
             gint width,
             gint height,
             GtkOrientation orientation)
{
  gint num_lines;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  gtk_draw_box (style, window, state_type, shadow_type, x, y,
                width, height);

  if (DETAIL ("scrollbar"))
    num_lines = 3;
  else
    num_lines = 2;

  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    draw_slider_knurl (style, window, state_type, area, widget,
                       x + (width / 2), y + 2, height - 4,
                       num_lines, orientation);
  else
    draw_slider_knurl (style, window, state_type, area, widget,
                       x + 2, y + (height / 2), width - 4,
                       num_lines, orientation);
}

static void
draw_handle(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
  gint i;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  /* Ideally, I would like handles to prelight, but no-one codes 
     their widgets that way. */

  if (orientation == GTK_ORIENTATION_VERTICAL)
    for (i = x + 2; i < (x + width - 2); i += 3)
      draw_vline (style, window, state_type, area, widget, "handle",
                  y + 2, y + height - 3, i);
  else
    for (i = y + 2; i < (y + height - 2); i += 3)
      draw_hline (style, window, state_type, area, widget, "handle",
                  x + 2, x + width - 3, i);
}
